"""
Basic neural network layers for hyperbolic space.
"""

import math
from typing import Optional, Union

import torch
import torch.nn as nn
import torch.nn.init as init
from hyptorch.pmath.poincare import distance, exponential_map_at_zero, mobius_addition, mobius_matrix_vector_multiplication, project


class HypLinear(nn.Module):
    """
    Hyperbolic linear layer.

    Parameters
    ----------
    in_features : int
        Number of input features.
    out_features : int
        Number of output features.
    curvature : float
        Negative ball curvature.
    bias : bool
        Whether to use bias.
    """

    def __init__(self, in_features: int, out_features: int, curvature: float, bias: bool = True):
        super(HypLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.curvature = curvature
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        """
        Reset the parameters of the module.
        """
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def forward(
        self, x: torch.Tensor, curvature: Optional[Union[float, torch.Tensor]] = None
    ) -> torch.Tensor:
        """
        Forward pass for hyperbolic linear layer.

        Parameters
        ----------
        x : tensor
            Input tensor.
        curvature : float or tensor, optional
            Negative ball curvature. If None, uses the class attribute.

        Returns
        -------
        tensor
            Output tensor.
        """
        # Use provided curvature or fall back to class attribute
        if curvature is None:
            curvature = self.curvature

        # Apply mobius matrix-vector multiplication
        mv = mobius_matrix_vector_multiplication(self.weight, x, curvature=curvature)

        # Apply bias if provided
        if self.bias is None:
            return project(mv, curvature=curvature)
        else:
            # Map bias to Poincare ball
            bias = exponential_map_at_zero(self.bias, curvature=curvature)

            # Add bias and project back to manifold
            return project(mobius_addition(mv, bias, curvature=curvature), curvature=curvature)

    def extra_repr(self) -> str:
        """
        Return a string representation of module parameters.

        Returns
        -------
        str
            String representation.
        """
        return "in_features={}, out_features={}, bias={}, curvature={}".format(
            self.in_features, self.out_features, self.bias is not None, self.curvature
        )


class ConcatPoincareLayer(nn.Module):
    """
    Layer for concatenating two points in Poincare ball.

    Parameters
    ----------
    d1 : int
        Dimension of first input.
    d2 : int
        Dimension of second input.
    d_out : int
        Dimension of output.
    curvature : float
        Negative ball curvature.
    """

    def __init__(self, d1: int, d2: int, d_out: int, curvature: float):
        super(ConcatPoincareLayer, self).__init__()
        self.d1 = d1
        self.d2 = d2
        self.d_out = d_out

        # Create hyperbolic linear layers for each input
        self.l1 = HypLinear(d1, d_out, bias=False, curvature=curvature)
        self.l2 = HypLinear(d2, d_out, bias=False, curvature=curvature)
        self.curvature = curvature

    def forward(
        self,
        x1: torch.Tensor,
        x2: torch.Tensor,
        curvature: Optional[Union[float, torch.Tensor]] = None,
    ) -> torch.Tensor:
        """
        Forward pass for concatenation layer.

        Parameters
        ----------
        x1 : tensor
            First input tensor.
        x2 : tensor
            Second input tensor.
        curvature : float or tensor, optional
            Negative ball curvature. If None, uses the class attribute.

        Returns
        -------
        tensor
            Concatenated tensor in Poincare ball.
        """
        # Use provided curvature or fall back to class attribute
        if curvature is None:
            curvature = self.curvature

        # Transform inputs using hyperbolic linear layers
        l1_out = self.l1(x1, curvature=curvature)
        l2_out = self.l2(x2, curvature=curvature)

        # Combine using mobius addition
        return mobius_addition(l1_out, l2_out, curvature=curvature)

    def extra_repr(self) -> str:
        """
        Return a string representation of module parameters.

        Returns
        -------
        str
            String representation.
        """
        return "dims {} and {} ---> dim {}".format(self.d1, self.d2, self.d_out)


class HyperbolicDistanceLayer(nn.Module):
    """
    Layer for computing hyperbolic distance between points.

    Parameters
    ----------
    curvature : float
        Negative ball curvature.
    """

    def __init__(self, curvature: float):
        super(HyperbolicDistanceLayer, self).__init__()
        self.curvature = curvature

    def forward(
        self,
        x1: torch.Tensor,
        x2: torch.Tensor,
        curvature: Optional[Union[float, torch.Tensor]] = None,
    ) -> torch.Tensor:
        """
        Forward pass for distance layer.

        Parameters
        ----------
        x1 : tensor
            First input tensor.
        x2 : tensor
            Second input tensor.
        curvature : float or tensor, optional
            Negative ball curvature. If None, uses the class attribute.

        Returns
        -------
        tensor
            Hyperbolic distance between inputs.
        """
        # Use provided curvature or fall back to class attribute
        if curvature is None:
            curvature = self.curvature

        # Compute hyperbolic distance
        return distance(x1, x2, curvature=curvature, keepdim=True)

    def extra_repr(self) -> str:
        """
        Return a string representation of module parameters.

        Returns
        -------
        str
            String representation.
        """
        return "curvature={}".format(self.curvature)



"""
Operations for conversions between models.
"""

from typing import Union

import torch


def poincare_to_klein(x: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Map from Poincare to Klein model.

    Parameters
    ----------
    x : tensor
        Point on Poincare ball.
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Point in Klein model.
    """
    c = torch.as_tensor(curvature).type_as(x)
    denominator = 1 + c * x.pow(2).sum(-1, keepdim=True)
    return 2 * x / denominator


def klein_to_poincare(x: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Map from Klein to Poincare model.

    Parameters
    ----------
    x : tensor
        Point in Klein model.
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Point on Poincare ball.
    """
    c = torch.as_tensor(curvature).type_as(x)
    denominator = 1 + torch.sqrt(1 - c * x.pow(2).sum(-1, keepdim=True))
    return x / denominator



"""
Custom autograd functions for hyperbolic operations.
"""

import torch

from ..config import CLAMP_MAX, CLAMP_MIN, EPS


class Artanh(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        x = x.clamp(CLAMP_MIN, CLAMP_MAX)
        ctx.save_for_backward(x)
        res = (torch.log_(1 + x).sub_(torch.log_(1 - x))).mul_(0.5)
        return res

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        return grad_output / (1 - input**2)


class Arsinh(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return (x + torch.sqrt_(1 + x.pow(2))).clamp_min_(EPS).log_()

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        return grad_output / (1 + input**2) ** 0.5


class RiemannianGradient(torch.autograd.Function):
    # TODO: Add curvature as a parameter
    curvature = 1

    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return x

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        # Compute Riemannian gradient scale factor
        scale = (1 - RiemannianGradient.curvature * input.pow(2).sum(-1, keepdim=True)).pow(2) / 4
        return grad_output * scale


def artanh(x: torch.Tensor) -> torch.Tensor:
    """
    Inverse hyperbolic tangent function.

    Parameters
    ----------
    x : tensor
        Input tensor.

    Returns
    -------
    tensor
        Inverse hyperbolic tangent of the input tensor.
    """
    return Artanh.apply(x)


def arsinh(x: torch.Tensor) -> torch.Tensor:
    """
    Inverse hyperbolic sine function.

    Parameters
    ----------
    x : tensor
        Input tensor.

    Returns
    -------
    tensor
        Inverse hyperbolic sine of the input tensor.
    """
    return Arsinh.apply(x)


def arcosh(x: torch.Tensor, eps: float = EPS) -> torch.Tensor:
    """
    Inverse hyperbolic cosine function.

    Parameters
    ----------
    x : tensor
        Input tensor.
    eps : float
        Epsilon for numerical stability.

    Returns
    -------
    tensor
        Inverse hyperbolic cosine of the input tensor.
    """
    x = x.clamp(-1 + eps, 1 - eps)
    return torch.log(x + torch.sqrt(1 + x) * torch.sqrt(x - 1))



"""
Distance computations in hyperbolic space.
"""

from typing import Union

import numpy as np
import torch

from scipy.special import gamma
from hyptorch.pmath.autograd import artanh
from hyptorch.pmath.poincare import mobius_addition_batch


def distance_matrix(x: torch.Tensor, y: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Compute distance matrix between two sets of points.

    Parameters
    ----------
    x : tensor
        First set of points.
    y : tensor
        Second set of points.
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Distance matrix.
    """
    c = torch.as_tensor(curvature).type_as(x)
    sqrt_curvature = c**0.5

    mobius_addition_result = mobius_addition_batch(-x, y, curvature=c)
    norm = torch.norm(mobius_addition_result, dim=-1)

    return 2 / sqrt_curvature * artanh(sqrt_curvature * norm)


def auto_select_c(dimension: int) -> float:
    """
    Calculate the radius of the Poincare ball such that the d-dimensional ball has constant volume equal to pi.

    Parameters
    ----------
    d : int
        Dimension of the ball.

    Returns
    -------
    float
        Computed curvature.
    """
    dim2 = dimension / 2.0
    R = gamma(dim2 + 1) / (np.pi ** (dim2 - 1))
    R = R ** (1 / float(dimension))
    curvature = 1 / (R**2)
    return curvature



"""
Core operations in the Poincaré ball model of hyperbolic space.
"""

from typing import Union

import torch
from pmath.autograd import arsinh, artanh
from utils.numeric import safe_tanh

from hyptorch.config import EPS
from hyptorch.utils.numeric import safe_tanh
from hyptorch.pmath.mappings import klein_to_poincare, poincare_to_klein


def project(x: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Safe projection on the manifold for numerical stability.

    Parameters
    ----------
    x : tensor
        Point on the Poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        Projected vector on the manifold.
    """
    c = torch.as_tensor(curvature).type_as(x)
    norm = torch.clamp_min(x.norm(dim=-1, keepdim=True, p=2), EPS)
    # TODO: Used to be EPS=1e-3, just in case of numerical instability
    maxnorm = (1 - EPS) / (c**0.5)
    cond = norm > maxnorm
    projected = x / norm * maxnorm
    return torch.where(cond, projected, x)


def compute_conformal_factor(
    x: torch.Tensor, curvature: Union[float, torch.Tensor], *, keepdim: bool = False
) -> torch.Tensor:
    """
    Compute the conformal factor for a point on the ball.

    Parameters
    ----------
    x : tensor
        Point on the Poincare ball.
    curvature : float or tensor
        Ball negative curvature.
    keepdim : bool
        Retain the last dim? (default: false)

    Returns
    -------
    tensor
        Conformal factor.
    """
    c = torch.as_tensor(curvature).type_as(x)
    return 2 / (1 - c * x.pow(2).sum(-1, keepdim=keepdim))


def mobius_addition(x: torch.Tensor, y: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Mobius addition in hyperbolic space.
    In general, this operation is not commutative.

    Parameters
    ----------
    x : tensor
        Point on the Poincare ball.
    y : tensor
        Point on the Poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        The result of mobius addition.
    """
    c = torch.as_tensor(curvature).type_as(x)

    # x2 and y2 are the squared norms of x and y
    # xy is the dot product of x and y
    x2 = x.pow(2).sum(dim=-1, keepdim=True)
    y2 = y.pow(2).sum(dim=-1, keepdim=True)
    xy = (x * y).sum(dim=-1, keepdim=True)

    numerator = (1 + 2 * c * xy + c * y2) * x + (1 - c * x2) * y
    denominator = 1 + 2 * c * xy + c**2 * x2 * y2

    return numerator / (denominator + EPS)


def mobius_matrix_vector_multiplication(
    matrix: torch.Tensor, x: torch.Tensor, curvature: Union[float, torch.Tensor]
) -> torch.Tensor:
    """
    Generalization for matrix-vector multiplication to hyperbolic space.

    Parameters
    ----------
    m : tensor
        Matrix for multiplication.
    x : tensor
        Point on poincare ball.
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Mobius matvec result.
    """
    c = torch.as_tensor(curvature).type_as(x)

    x_norm = torch.clamp_min(x.norm(dim=-1, keepdim=True, p=2), EPS)
    sqrt_c = c**0.5

    mx = x @ matrix.transpose(-1, -2)
    mx_norm = mx.norm(dim=-1, keepdim=True, p=2)

    res_c = safe_tanh(mx_norm / x_norm * artanh(sqrt_c * x_norm)) * mx / (mx_norm * sqrt_c)

    cond = (mx == 0).prod(-1, keepdim=True, dtype=torch.uint8)
    res_0 = torch.zeros(1, dtype=res_c.dtype, device=res_c.device)
    res = torch.where(cond, res_0, res_c)

    return project(res, curvature=c)


def exponential_map_at_zero(
    tangent_vector: torch.Tensor, curvature: Union[float, torch.Tensor]
) -> torch.Tensor:
    """
    Exponential map for Poincare ball model from 0.

    Parameters
    ----------
    tangent_vector : tensor
        Speed vector on poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        End point gamma_{0,tangent_vector}(1).
    """
    c = torch.as_tensor(curvature).type_as(tangent_vector)
    sqrt_c = c**0.5

    tangent_norm = torch.clamp_min(tangent_vector.norm(dim=-1, keepdim=True, p=2), EPS)

    return safe_tanh(sqrt_c * tangent_norm) * tangent_vector / (sqrt_c * tangent_norm)


def exponential_map(
    x: torch.Tensor, tangent_vector: torch.Tensor, curvature: Union[float, torch.Tensor]
) -> torch.Tensor:
    """
    Exponential map for Poincare ball model.

    Parameters
    ----------
    x : tensor
        Starting point on poincare ball.
    tangent_vector : tensor
        Speed vector on poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        End point gamma_{x,tangent_vector}(1).
    """
    c = torch.as_tensor(curvature).type_as(x)
    sqrt_c = c**0.5

    tangent_norm = torch.clamp_min(tangent_vector.norm(dim=-1, keepdim=True, p=2), EPS)

    # Calculate lambda_x(x, curvature)
    conformal_factor = compute_conformal_factor(x, curvature=c, keepdim=True)

    # Calculate second term
    second_term = (
        safe_tanh(sqrt_c / 2 * conformal_factor * tangent_norm) * tangent_vector / (sqrt_c * tangent_norm)
    )

    # Calculate result using mobius addition
    return mobius_addition(x, second_term, curvature=c)


def logarithmic_map_at_zero(y: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Logarithmic map for y from 0 on the manifold.

    Parameters
    ----------
    y : tensor
        Target point on poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        Tangent vector that transports 0 to y.
    """
    c = torch.as_tensor(curvature).type_as(y)
    sqrt_c = c**0.5

    y_norm = torch.clamp_min(y.norm(dim=-1, p=2, keepdim=True), EPS)

    return y / y_norm / sqrt_c * artanh(sqrt_c * y_norm)


def logarithmic_map(x: torch.Tensor, y: torch.Tensor, curvature: Union[float, torch.Tensor]) -> torch.Tensor:
    """
    Logarithmic map for two points x and y on the manifold.

    Parameters
    ----------
    x : tensor
        Starting point on poincare ball.
    y : tensor
        Target point on poincare ball.
    curvature : float or tensor
        Ball negative curvature.

    Returns
    -------
    tensor
        Tangent vector that transports x to y.
    """
    c = torch.as_tensor(curvature).type_as(x)

    # Calculate -x ⊕_c y
    difference_vector = mobius_addition(-x, y, curvature=c)

    difference_norm = difference_vector.norm(dim=-1, p=2, keepdim=True)

    # Calculate lambda_x(x, curvature)
    conformal_factor = compute_conformal_factor(x, curvature=c, keepdim=True)

    sqrt_c = c**0.5

    return (
        2 / sqrt_c / conformal_factor * artanh(sqrt_c * difference_norm) * difference_vector / difference_norm
    )


def distance(
    x: torch.Tensor,
    y: torch.Tensor,
    curvature: Union[float, torch.Tensor],
    *,
    keepdim: bool = False,
) -> torch.Tensor:
    """
    Distance on the Poincare ball.

    Parameters
    ----------
    x : tensor
        Point on poincare ball.
    y : tensor
        Point on poincare ball.
    curvature : float or tensor
        Ball negative curvature.
    keepdim : bool
        Retain the last dim? (default: false)

    Returns
    -------
    tensor
        Geodesic distance between x and y.
    """
    c = torch.as_tensor(curvature).type_as(x)
    sqrt_c = c**0.5

    dist_c = artanh(sqrt_c * mobius_addition(-x, y, c).norm(dim=-1, p=2, keepdim=keepdim))

    return dist_c * 2 / sqrt_c


def distance_from_center(
    x: torch.Tensor,
    curvature: Union[float, torch.Tensor],
    *,
    keepdim: bool = False,
) -> torch.Tensor:
    """
    Distance on the Poincare ball to zero.

    Parameters
    ----------
    x : tensor
        Point on poincare ball.
    curvature : float or tensor
        Ball negative curvature.
    keepdim : bool
        Retain the last dim? (default: false)

    Returns
    -------
    tensor
        Geodesic distance between x and 0.
    """
    c = torch.as_tensor(curvature).type_as(x)
    sqrt_c = c**0.5
    dist_c = artanh(sqrt_c * x.norm(dim=-1, p=2, keepdim=keepdim))

    return dist_c * 2 / sqrt_c


def poincare_mean(
    x: torch.Tensor,
    curvature: Union[float, torch.Tensor],
    dim: int = 0,
) -> torch.Tensor:
    """
    Compute mean in Poincare ball model.

    Parameters
    ----------
    x : tensor
        Points on Poincare ball.
    curvature : float or tensor
        Negative ball curvature.
    dim : int
        Dimension along which to compute mean.

    Returns
    -------
    tensor
        Mean in Poincare ball model.
    """
    # Convert to Klein model
    x_klein = poincare_to_klein(x, curvature)

    # Compute Lorenz factor
    lamb = lorenz_factor(x_klein, curvature=curvature, keepdim=True)

    # Compute weighted sum
    lamb_sum = torch.sum(lamb, dim=dim, keepdim=True)
    weighted_sum = torch.sum(lamb * x_klein, dim=dim, keepdim=True) / lamb_sum

    # Convert back to Poincare ball
    mean_poincare = klein_to_poincare(weighted_sum, curvature)

    return mean_poincare.squeeze(dim)


def lorenz_factor(
    x: torch.Tensor,
    curvature: Union[float, torch.Tensor],
    *,
    dim: int = -1,
    keepdim: bool = False,
) -> torch.Tensor:
    """
    Compute Lorenz factor.

    Parameters
    ----------
    x : tensor
        Point on Klein disk.
    curvature : float
        Negative curvature.
    dim : int
        Dimension to calculate Lorenz factor.
    keepdim : bool
        Retain the last dim? (default: false)

    Returns
    -------
    tensor
        Lorenz factor.
    """
    return 1 / torch.sqrt(1 - curvature * x.pow(2).sum(dim=dim, keepdim=keepdim))


def mobius_addition_batch(
    x: torch.Tensor, y: torch.Tensor, curvature: Union[float, torch.Tensor]
) -> torch.Tensor:
    """
    Compute mobius addition in batch mode.

    Parameters
    ----------
    x : tensor
        First tensor (batch of points).
    y : tensor
        Second tensor (batch of points).
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Batch mobius addition result.
    """
    xy = torch.einsum("ij,kj->ik", (x, y))  # B x C
    x2 = x.pow(2).sum(-1, keepdim=True)  # B x 1
    y2 = y.pow(2).sum(-1, keepdim=True)  # C x 1

    num = 1 + 2 * curvature * xy + curvature * y2.permute(1, 0)  # B x C
    num = num.unsqueeze(2) * x.unsqueeze(1)
    num = num + (1 - curvature * x2).unsqueeze(2) * y  # B x C x D

    denom_part1 = 1 + 2 * curvature * xy  # B x C
    denom_part2 = curvature**2 * x2 * y2.permute(1, 0)
    denom = denom_part1 + denom_part2

    return num / (denom.unsqueeze(2) + EPS)


def hyperbolic_softmax(
    X: torch.Tensor,
    A: torch.Tensor,
    P: torch.Tensor,
    curvature: Union[float, torch.Tensor],
) -> torch.Tensor:
    """
    Compute hyperbolic softmax.

    Parameters
    ----------
    X : tensor
        Input tensor.
    A : tensor
        Weights tensor.
    P : tensor
        Points tensor.
    curvature : float or tensor
        Negative ball curvature.

    Returns
    -------
    tensor
        Hyperbolic softmax result.
    """
    # Pre-compute common values
    lambda_pkc = 2 / (1 - curvature * P.pow(2).sum(dim=1))
    k = lambda_pkc * torch.norm(A, dim=1) / torch.sqrt(curvature)

    # Calculate mobius addition and other values
    mob_add = mobius_addition_batch(-P, X, curvature)

    num = 2 * torch.sqrt(curvature) * torch.sum(mob_add * A.unsqueeze(1), dim=-1)

    denom = torch.norm(A, dim=1, keepdim=True) * (1 - curvature * mob_add.pow(2).sum(dim=2))

    logit = k.unsqueeze(1) * arsinh(num / denom)

    return logit.permute(1, 0)



"""
Functional operations for hyperbolic neural networks.
"""

import torch

from hyptorch.pmath.poincare import hyperbolic_softmax


def hyperbolic_softmax_loss(
    logits: torch.Tensor,
    targets: torch.Tensor,
    weights: torch.Tensor,
    points: torch.Tensor,
    curvature: float,
    reduction: str = "mean",
) -> torch.Tensor:
    """
    Softmax loss in hyperbolic space.

    Parameters
    ----------
    logits : tensor
        Predicted logits from the model.
    targets : tensor
        Target labels.
    points : tensor
        Points in Poincaré ball for hyperbolic softmax.
    curvature : float
        Negative ball curvature.
    reduction : str
        Reduction method ('mean', 'sum', or 'none').

    Returns
    -------
    tensor
        Loss value.
    """
    probs = hyperbolic_softmax(logits, weights, points, curvature)
    log_probs = torch.log(probs + 1e-8)

    if reduction == "none":
        return -log_probs.gather(1, targets.unsqueeze(-1))
    elif reduction == "sum":
        return -log_probs.gather(1, targets.unsqueeze(-1)).sum()
    else:  # mean
        return -log_probs.gather(1, targets.unsqueeze(-1)).mean()
